import{_ as v,A as o,E as u}from"./index.992e7f6f.js";import{e as w,M as y,ae as k,s,D as h,r as C,o as d,m as _,w as g,d as b,u as a,g as B}from"./vendor.9c8de484.js";import{T as R}from"./Timeline.57f70d80.js";const T={class:"timeline-content py-2 px-2 d-flex flex-grow-1"},q=w({async setup(j){let t,r;const i=y(),n=([t,r]=k(()=>o.query({endpoint:u.flow_run,body:{id:i.params.id},options:{pollInterval:5e3}}).fetch()),t=await t,r(),t),c=s(()=>({flow_runs:{id:{any_:[i.params.id]}}})),p={task_runs:o.query({endpoint:u.task_runs,body:c,options:{pollInterval:1e4}})},m=s(()=>{var e;return((e=n.response)==null?void 0:e.value)||{}}),f=s(()=>m.value.end_time),l=s(()=>{var e;return((e=p.task_runs.response)==null?void 0:e.value)||[]});return h(()=>{n.stopPolling(),o.queries.delete(n.id)}),(e,A)=>{const x=C("Card");return d(),_(x,{class:"timeline d-flex flex-column",width:"auto",shadow:"sm"},{default:g(()=>[b("div",T,[a(l).length?(d(),_(R,{key:0,items:a(l),"max-end-time":a(f),"axis-position":"top","background-color":"blue-5"},null,8,["items","max-end-time"])):B("",!0)])]),_:1})}}});var V=v(q,[["__scopeId","data-v-67b91c6e"]]);export{V as default};
