import{_ as o}from"./index.992e7f6f.js";import{r as c,o as a,m as i,w as l,p as r,b as n,d as e}from"./vendor.9c8de484.js";const d={},p=t=>(r("data-v-44636fc7"),t=t(),n(),t),_=p(()=>e("div",{class:"schematic-content pa-2 d-flex flex-grow-1"},[e("div",{style:{top:"50%",left:"50%",transform:"translate(-50%, -50%)",position:"absolute"},class:"text-center d-flex align-center"},[e("i",{class:"pi pi-tools-fill pi-2x"}),e("h1",{class:"mx-2"},"Under construction"),e("i",{class:"pi pi-tools-fill pi-2x"})])],-1));function f(t,m){const s=c("Card");return a(),i(s,{class:"schematic d-flex flex-column",width:"auto",shadow:"sm"},{default:l(()=>[_]),_:1})}var u=o(d,[["render",f],["__scopeId","data-v-44636fc7"]]);export{u as default};
