.shadow-sm[data-v-67b91c6e]{box-shadow:0 1px 2px #0000000f,0 1px 3px #0000001a}.shadow-md[data-v-67b91c6e]{box-shadow:0 2px 4px #0000000f,0 4px 6px #0000001a}.shadow-lg[data-v-67b91c6e]{box-shadow:0 10px 15px #0000001a,0 4px 6px #0000000d}.shadow-xl[data-v-67b91c6e]{box-shadow:0 20px 25px #0000001a,0 10px 10px #0000000a}.shadow-2xl[data-v-67b91c6e]{box-shadow:0 25px 50px #00000040}.shadow-inset[data-v-67b91c6e]{box-shadow:inset 0 2px 4px #0000000f}.drop-shadow-sm[data-v-67b91c6e]{filter:drop-shadow(0px 1px 2px rgba(0,0,0,.06)) drop-shadow(0px 1px 3px rgba(0,0,0,.1))}.drop-shadow-md[data-v-67b91c6e]{filter:drop-shadow(0px 2px 4px rgba(0,0,0,.06)) drop-shadow(0px 4px 6px rgba(0,0,0,.1))}.drop-shadow-lg[data-v-67b91c6e]{filter:drop-shadow(0px 10px 15px rgba(0,0,0,.1)) drop-shadow(0px 4px 6px rgba(0,0,0,.05))}.drop-shadow-xl[data-v-67b91c6e]{filter:drop-shadow(0px 20px 25px rgba(0,0,0,.1)) drop-shadow(0px 10px 10px rgba(0,0,0,.04))}.drop-shadow-2xl[data-v-67b91c6e]{filter:drop-shadow(0px 25px 50px rgba(0,0,0,.25))}.timeline[data-v-67b91c6e]{height:80vh;min-height:600px;width:100%}.timeline .timeline-content[data-v-67b91c6e]{box-sizing:content-box;max-height:calc(80vh - 48px)}
