.shadow-sm{box-shadow:0 1px 2px #0000000f,0 1px 3px #0000001a}.shadow-md{box-shadow:0 2px 4px #0000000f,0 4px 6px #0000001a}.shadow-lg{box-shadow:0 10px 15px #0000001a,0 4px 6px #0000000d}.shadow-xl{box-shadow:0 20px 25px #0000001a,0 10px 10px #0000000a}.shadow-2xl{box-shadow:0 25px 50px #00000040}.shadow-inset{box-shadow:inset 0 2px 4px #0000000f}.drop-shadow-sm{filter:drop-shadow(0px 1px 2px rgba(0,0,0,.06)) drop-shadow(0px 1px 3px rgba(0,0,0,.1))}.drop-shadow-md{filter:drop-shadow(0px 2px 4px rgba(0,0,0,.06)) drop-shadow(0px 4px 6px rgba(0,0,0,.1))}.drop-shadow-lg{filter:drop-shadow(0px 10px 15px rgba(0,0,0,.1)) drop-shadow(0px 4px 6px rgba(0,0,0,.05))}.drop-shadow-xl{filter:drop-shadow(0px 20px 25px rgba(0,0,0,.1)) drop-shadow(0px 10px 10px rgba(0,0,0,.04))}.drop-shadow-2xl{filter:drop-shadow(0px 25px 50px rgba(0,0,0,.25))}.not-found{display:flex;align-items:center;justify-content:center;background:url(/ui/assets/not-found-bg-top.8502b54f.svg) top left no-repeat,url(/ui/assets/not-found-bg-bottom.01506542.svg) bottom right no-repeat,#0a2053}@media screen and (max-width: 1024px){.not-found{background-size:150px auto,500px auto}}.not-found__message{width:600px;max-width:80%;margin-top:-15%}.not-found__circle{transform-box:fill-box;transform-origin:center;opacity:0}.not-found__circle--animate{animation:pulse-black 1.5s}@keyframes pulse-black{0%{transform:scale(0);opacity:0}50%{opacity:1}to{transform:scale(1);opacity:0}}
