.shadow-sm[data-v-1e41a11a]{box-shadow:0 1px 2px #0000000f,0 1px 3px #0000001a}.shadow-md[data-v-1e41a11a]{box-shadow:0 2px 4px #0000000f,0 4px 6px #0000001a}.shadow-lg[data-v-1e41a11a]{box-shadow:0 10px 15px #0000001a,0 4px 6px #0000000d}.shadow-xl[data-v-1e41a11a]{box-shadow:0 20px 25px #0000001a,0 10px 10px #0000000a}.shadow-2xl[data-v-1e41a11a]{box-shadow:0 25px 50px #00000040}.shadow-inset[data-v-1e41a11a]{box-shadow:inset 0 2px 4px #0000000f}.drop-shadow-sm[data-v-1e41a11a]{filter:drop-shadow(0px 1px 2px rgba(0,0,0,.06)) drop-shadow(0px 1px 3px rgba(0,0,0,.1))}.drop-shadow-md[data-v-1e41a11a]{filter:drop-shadow(0px 2px 4px rgba(0,0,0,.06)) drop-shadow(0px 4px 6px rgba(0,0,0,.1))}.drop-shadow-lg[data-v-1e41a11a]{filter:drop-shadow(0px 10px 15px rgba(0,0,0,.1)) drop-shadow(0px 4px 6px rgba(0,0,0,.05))}.drop-shadow-xl[data-v-1e41a11a]{filter:drop-shadow(0px 20px 25px rgba(0,0,0,.1)) drop-shadow(0px 10px 10px rgba(0,0,0,.04))}.drop-shadow-2xl[data-v-1e41a11a]{filter:drop-shadow(0px 25px 50px rgba(0,0,0,.25))}.run-state[data-v-1e41a11a]{border-radius:4px;padding:2px 8px;text-transform:capitalize}.result-badge[data-v-1e41a11a]{background-color:#fff;border-radius:16px;font-weight:400;padding:0 8px;min-width:24px;transition:.15s all}.result-badge.active[data-v-1e41a11a]{background-color:#024dfd;color:#fff}.main-grid[data-v-1e41a11a]{display:grid;grid-template-areas:"details details schematic" "timeline timeline schematic" "timeline timeline schematic" "timeline timeline schematic";grid-template-columns:repeat(3,minmax(0,1fr));grid-template-rows:repeat(4,70px);column-gap:16px;row-gap:16px}.main-grid .details[data-v-1e41a11a],.main-grid .timeline[data-v-1e41a11a],.main-grid .schematic[data-v-1e41a11a]{height:100%}.main-grid .details[data-v-1e41a11a]{grid-area:details}.main-grid .timeline[data-v-1e41a11a]{grid-area:timeline}.main-grid .schematic[data-v-1e41a11a]{grid-area:schematic}.timeline .timeline-content[data-v-1e41a11a]{box-sizing:content-box;max-height:194px}.timeline.full-screen[data-v-1e41a11a]{height:60vh}.timeline.full-screen .timeline-content[data-v-1e41a11a]{max-height:calc(60vh - 40px)}.results-hr[data-v-1e41a11a]{background-color:#cacccf;border:none;height:2px}@media (max-width: 1024px){.main-grid[data-v-1e41a11a]{grid-template-areas:"details" "schematic" "timeline";grid-template-columns:repeat(1,minmax(0,1fr));grid-template-rows:70px 200px 250px}}
