.shadow-sm[data-v-9ef185d0]{box-shadow:0 1px 2px #0000000f,0 1px 3px #0000001a}.shadow-md[data-v-9ef185d0]{box-shadow:0 2px 4px #0000000f,0 4px 6px #0000001a}.shadow-lg[data-v-9ef185d0]{box-shadow:0 10px 15px #0000001a,0 4px 6px #0000000d}.shadow-xl[data-v-9ef185d0]{box-shadow:0 20px 25px #0000001a,0 10px 10px #0000000a}.shadow-2xl[data-v-9ef185d0]{box-shadow:0 25px 50px #00000040}.shadow-inset[data-v-9ef185d0]{box-shadow:inset 0 2px 4px #0000000f}.drop-shadow-sm[data-v-9ef185d0]{filter:drop-shadow(0px 1px 2px rgba(0,0,0,.06)) drop-shadow(0px 1px 3px rgba(0,0,0,.1))}.drop-shadow-md[data-v-9ef185d0]{filter:drop-shadow(0px 2px 4px rgba(0,0,0,.06)) drop-shadow(0px 4px 6px rgba(0,0,0,.1))}.drop-shadow-lg[data-v-9ef185d0]{filter:drop-shadow(0px 10px 15px rgba(0,0,0,.1)) drop-shadow(0px 4px 6px rgba(0,0,0,.05))}.drop-shadow-xl[data-v-9ef185d0]{filter:drop-shadow(0px 20px 25px rgba(0,0,0,.1)) drop-shadow(0px 10px 10px rgba(0,0,0,.04))}.drop-shadow-2xl[data-v-9ef185d0]{filter:drop-shadow(0px 25px 50px rgba(0,0,0,.25))}.component-container[data-v-9ef185d0]{overflow:hidden;height:100%;position:relative;width:100%}.component-container.header-visible[data-v-9ef185d0]{--axis-height: 38px}.component-container.header-hidden[data-v-9ef185d0]{--axis-height: 20px}.component-container .pan-button[data-v-9ef185d0]{position:absolute;z-index:3;top:0}.component-container .pan-button.left[data-v-9ef185d0]{left:0}.component-container .pan-button.right[data-v-9ef185d0]{right:0}.component-container .timeline-axis[data-v-9ef185d0]{background-color:#fff;height:var(--axis-height);pointer-events:none;position:sticky;top:0;left:0;z-index:2}.component-container svg.timeline-background[data-v-9ef185d0]{position:absolute;height:100%;left:0;width:100%;z-index:1}.component-container.axis-bottom svg.timeline-background[data-v-9ef185d0]{top:0}.component-container.axis-top svg.timeline-background[data-v-9ef185d0]{top:var(--axis-height)}.component-container .scroll-container[data-v-9ef185d0]{clip-path:inset(0% 0% 0% 0% round 4px);overflow:auto;overscroll-behavior:contain;position:relative;max-height:100%;height:100%;top:0;left:0;flex:1 1 auto;scrollbar-width:none;-ms-overflow-style:none}.component-container .scroll-container[data-v-9ef185d0]::-webkit-scrollbar{display:none}.component-container .border-container[data-v-9ef185d0]{border:1px solid #ebeef7;border-radius:4px;box-sizing:border-box;height:calc(100% - var(--axis-height));left:0;position:absolute;width:100%;pointer-events:none;z-index:2}.component-container .border-container.bottom[data-v-9ef185d0]{top:0}.component-container .border-container.top[data-v-9ef185d0]{top:var(--axis-height)}.component-container .node-container[data-v-9ef185d0]{background-color:transparent;position:relative;z-index:1}.component-container .node-container .node[data-v-9ef185d0]{position:absolute;border-radius:8px;transition:all .25s;z-index:2}.component-container .node-container .node[data-v-9ef185d0]:focus,.component-container .node-container .node[data-v-9ef185d0]:hover{border:2px solid #024dfd;box-shadow:0 2px 4px #0000000f,0 4px 6px #0000001a;outline:none}.shadow-sm{box-shadow:0 1px 2px #0000000f,0 1px 3px #0000001a}.shadow-md{box-shadow:0 2px 4px #0000000f,0 4px 6px #0000001a}.shadow-lg{box-shadow:0 10px 15px #0000001a,0 4px 6px #0000000d}.shadow-xl{box-shadow:0 20px 25px #0000001a,0 10px 10px #0000000a}.shadow-2xl{box-shadow:0 25px 50px #00000040}.shadow-inset{box-shadow:inset 0 2px 4px #0000000f}.drop-shadow-sm{filter:drop-shadow(0px 1px 2px rgba(0,0,0,.06)) drop-shadow(0px 1px 3px rgba(0,0,0,.1))}.drop-shadow-md{filter:drop-shadow(0px 2px 4px rgba(0,0,0,.06)) drop-shadow(0px 4px 6px rgba(0,0,0,.1))}.drop-shadow-lg{filter:drop-shadow(0px 10px 15px rgba(0,0,0,.1)) drop-shadow(0px 4px 6px rgba(0,0,0,.05))}.drop-shadow-xl{filter:drop-shadow(0px 20px 25px rgba(0,0,0,.1)) drop-shadow(0px 10px 10px rgba(0,0,0,.04))}.drop-shadow-2xl{filter:drop-shadow(0px 25px 50px rgba(0,0,0,.25))}.timeline-axis>svg .tick{font-size:13px;font-family:"input-sans"}.timeline-axis>svg .domain{opacity:0}.timeline-axis>svg>g{transform:translateY(25%)}
