import{_ as B,A as u,E as m}from"./index.992e7f6f.js";import{e as R,M as q,f as D,ae as P,s as _,D as S,H as A,r as w,y as C,o as F,c as V,d as s,a as v,u as h,t as j,A as E,q as b,p as N,b as T}from"./vendor.9c8de484.js";const M=t=>(N("data-v-3deb768a"),t=t(),T(),t),H={class:"d-flex align-center justify-space-between mb-2"},U={class:"text-truncate"},$=b(" Flow Version: "),z={class:"font-weight-semibold"},G={class:"copy-link ml-1"},J=M(()=>s("i",{class:"pi pi-link pi-xs"},null,-1)),K=b(" Copy Run ID "),L=[J,K],O=R({async setup(t){let o,d;const n=q(),x=D(null),p=n.params.id,a=([o,d]=P(()=>u.query({endpoint:m.flow_run,body:{id:p},options:{pollInterval:5e3}}).fetch()),o=await o,d(),o),y={id:a.response.value.flow_id},g={flow:u.query({endpoint:m.flow,body:y})},I=_(()=>{var e;return((e=g.flow.response)==null?void 0:e.value)||{}}),f=_(()=>{var e;return((e=a.response)==null?void 0:e.value)||{}}),k=_(()=>{var i,r;const e=[{text:(i=I.value)==null?void 0:i.name},{text:(r=f.value)==null?void 0:r.name,to:""}],l=n.fullPath.includes("/timeline"),c=n.fullPath.includes("/schematic");return(l||c)&&(e[1].to=`/flow-run/${p}`,l&&e.push({text:"Timeline"}),c&&e.push({text:"Schematic"})),e});return S(()=>{a.stopPolling(),u.queries.delete(a.id)}),A(()=>{var e;x.value=((e=n.hash)==null?void 0:e.substr(1))||"task_runs"}),(e,l)=>{const c=w("bread-crumbs"),i=w("router-view"),r=C("breakpoints");return F(),V("div",null,[s("div",H,[v(c,{class:"flex-grow-1",crumbs:h(k),icon:"pi-flow-run"},null,8,["crumbs"]),s("div",U,[s("span",null,[$,s("span",z,j(h(f).flow_version),1)]),E(s("a",G,L,512),[[r,"md"]])])]),v(i)])}}});var Y=B(O,[["__scopeId","data-v-3deb768a"]]);export{Y as default};
