.shadow-sm[data-v-9fc654de]{box-shadow:0 1px 2px #0000000f,0 1px 3px #0000001a}.shadow-md[data-v-9fc654de]{box-shadow:0 2px 4px #0000000f,0 4px 6px #0000001a}.shadow-lg[data-v-9fc654de]{box-shadow:0 10px 15px #0000001a,0 4px 6px #0000000d}.shadow-xl[data-v-9fc654de]{box-shadow:0 20px 25px #0000001a,0 10px 10px #0000000a}.shadow-2xl[data-v-9fc654de]{box-shadow:0 25px 50px #00000040}.shadow-inset[data-v-9fc654de]{box-shadow:inset 0 2px 4px #0000000f}.drop-shadow-sm[data-v-9fc654de]{filter:drop-shadow(0px 1px 2px rgba(0,0,0,.06)) drop-shadow(0px 1px 3px rgba(0,0,0,.1))}.drop-shadow-md[data-v-9fc654de]{filter:drop-shadow(0px 2px 4px rgba(0,0,0,.06)) drop-shadow(0px 4px 6px rgba(0,0,0,.1))}.drop-shadow-lg[data-v-9fc654de]{filter:drop-shadow(0px 10px 15px rgba(0,0,0,.1)) drop-shadow(0px 4px 6px rgba(0,0,0,.05))}.drop-shadow-xl[data-v-9fc654de]{filter:drop-shadow(0px 20px 25px rgba(0,0,0,.1)) drop-shadow(0px 10px 10px rgba(0,0,0,.04))}.drop-shadow-2xl[data-v-9fc654de]{filter:drop-shadow(0px 25px 50px rgba(0,0,0,.25))}.theme-option[data-v-9fc654de]{width:100%}.theme-option .theme-palette[data-v-9fc654de]{display:inline-block;align-self:flex-end}.theme-option .theme-palette .palette-option[data-v-9fc654de]{border-radius:50%;border:1px solid #eee;display:inline-block;margin-right:4px;height:15px;width:15px}.shadow-sm[data-v-2cafc35e]{box-shadow:0 1px 2px #0000000f,0 1px 3px #0000001a}.shadow-md[data-v-2cafc35e]{box-shadow:0 2px 4px #0000000f,0 4px 6px #0000001a}.shadow-lg[data-v-2cafc35e]{box-shadow:0 10px 15px #0000001a,0 4px 6px #0000000d}.shadow-xl[data-v-2cafc35e]{box-shadow:0 20px 25px #0000001a,0 10px 10px #0000000a}.shadow-2xl[data-v-2cafc35e]{box-shadow:0 25px 50px #00000040}.shadow-inset[data-v-2cafc35e]{box-shadow:inset 0 2px 4px #0000000f}.drop-shadow-sm[data-v-2cafc35e]{filter:drop-shadow(0px 1px 2px rgba(0,0,0,.06)) drop-shadow(0px 1px 3px rgba(0,0,0,.1))}.drop-shadow-md[data-v-2cafc35e]{filter:drop-shadow(0px 2px 4px rgba(0,0,0,.06)) drop-shadow(0px 4px 6px rgba(0,0,0,.1))}.drop-shadow-lg[data-v-2cafc35e]{filter:drop-shadow(0px 10px 15px rgba(0,0,0,.1)) drop-shadow(0px 4px 6px rgba(0,0,0,.05))}.drop-shadow-xl[data-v-2cafc35e]{filter:drop-shadow(0px 20px 25px rgba(0,0,0,.1)) drop-shadow(0px 10px 10px rgba(0,0,0,.04))}.drop-shadow-2xl[data-v-2cafc35e]{filter:drop-shadow(0px 25px 50px rgba(0,0,0,.25))}@keyframes skeleton-loader-animation-2cafc35e{to{transform:translate(100%)}}.setting--section .setting--section-title[data-v-2cafc35e]{text-transform:capitalize}.setting--section .setting--section-content[data-v-2cafc35e]{--end: 100%}@-moz-document url-prefix(){.setting--section .setting--section-content[data-v-2cafc35e]{--end: calc(100% + .5px)}}@media not all and (min-resolution: .001dpcm){.setting--section .setting--section-content[data-v-2cafc35e]{--end: calc(100% + .5px)}}.setting--section .setting--section-content[data-v-2cafc35e]>:first-child{clip-path:stroke-box polygon(calc(var(--end) - 16px) var(--end),var(--end) calc(var(--end) - 16px),var(--end) 0,0 0,0 var(--end),calc(var(--end) - 16px) var(--end));-webkit-clip-path:polygon(calc(var(--end) - 16px) var(--end),var(--end) calc(var(--end) - 16px),var(--end) 0,0 0,0 var(--end),calc(var(--end) - 16px) var(--end));position:relative}.setting--section .setting--section-content[data-v-2cafc35e]>:first-child:after{content:"";height:23px;width:0;padding:0;right:14.75px;bottom:-1.25px;position:absolute;border-color:inherit;border:inherit;border-bottom:none;border-right:none;border-top:none;border-width:1.25px;transform:rotate(45deg);transform-origin:bottom;user-select:none;-webkit-user-select:none}.setting--section .setting--section-content>div[data-v-2cafc35e]{background-color:#f7f7f8;border:1px solid #cecdd3;border-radius:3px}.setting--section .setting--section-content section[data-v-2cafc35e],.setting--section .setting--section-content div[data-v-2cafc35e]{margin:4px 0}
